/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt;

import de.bsvrz.ars.ars.mgmt.ArchivConfig.EndOfParamListener;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveSettingsTask;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveSettingsTask.ArSSettingListener;
import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavParameters;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.CommunicationError;
import de.bsvrz.dav.daf.main.ConnectionException;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.InconsistentLoginException;
import de.bsvrz.dav.daf.main.MissingParameterException;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveInfoQueryResult;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.dataSerializer.Deserializer;
import de.bsvrz.sys.funclib.dataSerializer.Serializer;
import de.bsvrz.sys.funclib.dataSerializer.SerializingFactory;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.ArchiveSettings;
import de.bsvrz.sys.funclib.losb.datk.ContainerSettings;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.datk.PidQuery;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.SenderException;
import de.bsvrz.sys.funclib.losb.kernsoftware.SimpleSender;
import de.bsvrz.sys.funclib.losb.util.JavaSpy;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.util.*;
import java.util.logging.Level;

/**
 * Klasse zur Fernsteuerung des Archivsystems (fuer Testzwecke).
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 8003 $ / $Date: 2010-07-15 20:00:05 +0200 (Thu, 15 Jul 2010) $ / ($Author: rs $)
 */
public class ArSRemoteControl {

	/** Wenn wahr, gibt die RemoteControl keine eigenen Debugmeldungen aus */
	public static boolean DISABLE_RC_DEBUG = true;

	private static final String SEND_ASPECT = "asp.parameterVorgabe";

	private static final String ANFR_ASPECT = "asp.anfrage";

	private static ClientDavParameters davParams;

	private static ClientDavConnection davCon = null;

	private static Level lastDebugLevel;

	public static String PASSWD = "src/de/bsvrz/pua/prot/testenvironment/passwd";

	public static String DAV_ADDRESS = "localhost:8083";

	/** Beim Start des Archivsystem zu verwendendes Debuglevel fr Ausgabe in Datei. */
	private static String _debugLevelFile = "INFO";

	/**
	 * @return Liefert das beim Start des Archivsystem zu verwendendes Debuglevel fr Ausgabe in Datei.
	 */
	public static String getDebugLevelFile() {
		return _debugLevelFile;
	}

	/**
	 * Setzt das beim Start des Archivsystem zu verwendendes Debuglevel fr Ausgabe in Datei.
	 * @param debugLevelFile Das beim Start des Archivsystem zu verwendendes Debuglevel fr Ausgabe in Datei.
	 */
	public static void setDebugLevelFile(final String debugLevelFile) {
		_debugLevelFile = debugLevelFile;
	}

	private static void suspendDebug() throws Exception {
		if(DISABLE_RC_DEBUG) {
			lastDebugLevel = (Level)JavaSpy.getStaticFieldvalue(Debug.class, "_debugLevelStdErrText");
			Debug.setHandlerLevel("StdErr", Debug.OFF);
		}
	}

	private static void resumeDebug() {
		if(DISABLE_RC_DEBUG && lastDebugLevel != null) Debug.setHandlerLevel("StdErr", lastDebugLevel);
	}

	/**
	 * Startet das Archivsystem ohne auf die Uebernahme der Parametrierung zu warten.
	 *
	 * @param persistenceDir Persistenzverzeichnis
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArS(String persistenceDir, String debugLevel) throws Exception {
		ArchiveManager am = new ArchiveManager(
				new String[]{
						"-debugLevelStdErrText=" + debugLevel, "-debugLevelFileText=" + _debugLevelFile, "-persistenzVerzeichnis=" + persistenceDir, "-benutzer=Tester", "-authentifizierung=" + PASSWD,
						"-sicherungsModul=keydox"
				}
		);
		am.run();
		return am;
	}

	public static ArchiveManager startArSRemoveOldData(String persistenceDir, String debugLevel) throws Exception {
		if(!Util.deleteCreateNewDir(new File(persistenceDir))) {
			return null;
		}
		else {
			return startArS(persistenceDir, debugLevel);
		}
	}

	/**
	 * Startet das Archivsystem. Es wird gewartet, bis die Parametrierung vollstaendig uebernommen wurde. Debug-Level fuer stdout ist "INFO".
	 *
	 * @param persistenceDir Persistenzverzeichnis
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArSWaitForParams(String persistenceDir) throws Exception {
		return startArSWaitForParams(persistenceDir, "INFO", null);
	}

	/**
	 * Startet das Archivsystem unter dem lokalen KV. Es wird gewartet, bis die Parametrierung vollstaendig uebernommen wurde.
	 *
	 * @param persistenceDir Persistenzverzeichnis
	 * @param debugLevel     Debug-Level fuer stdout
	 * @param configAuth     Konfigurationsverantwortlicher oder null (dann wird der lokale KV verwendet)
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArSWaitForParams(String persistenceDir, String debugLevel) throws Exception {
		return startArSWaitForParams(persistenceDir, debugLevel, null);
	}

	/**
	 * Startet das Archivsystem. Es wird gewartet, bis die Parametrierung vollstaendig uebernommen wurde.
	 *
	 * @param persistenceDir Persistenzverzeichnis
	 * @param debugLevel     Debug-Level fuer stdout
	 * @param configAuth     Konfigurationsverantwortlicher oder null (dann wird der lokale KV verwendet)
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArSWaitForParams(String persistenceDir, String debugLevel, String configAuth) throws Exception {
		String[] params;
		String backupPath = Util.deleteCreateNewTempDir("backup");
		if(configAuth == null) {
			params = new String[]{
					"-debugLevelStdErrText=" + debugLevel, "-debugLevelFileText=" + _debugLevelFile, "-persistenzVerzeichnis=" + persistenceDir, "-datenverteiler=" + DAV_ADDRESS, "-benutzer=Tester",
					"-authentifizierung=" + PASSWD, "-sicherungsModul=filesystem", "-sicherungsVerzeichnis=" + backupPath
			};
		}
		else {
			params = new String[]{
					"-konfigurationsVerantwortlicher=" + configAuth, "-debugLevelStdErrText=" + debugLevel, "-debugLevelFileText=" + _debugLevelFile, "-persistenzVerzeichnis=" + persistenceDir,
					"-datenverteiler=" + DAV_ADDRESS, "-benutzer=Tester", "-authentifizierung=" + PASSWD, "-sicherungsModul=filesystem",
					"-sicherungsVerzeichnis=" + backupPath
			};
		}
		return startArsWaitForParamsWithArgs(params);
	}

	/**
	 * Startet das Archivsystem. Es wird gewartet, bis die Parametrierung vollstaendig uebernommen wurde.<br> Falls im Parameter backupModule "filesystem"
	 * spezifiziert wurde, sollte die Methode {@link #startArSWaitForParams(String,String,String,String,String,boolean)} verwendet werden und ein Verzeichnis fuer
	 * die Sicherung angegeben werden (ansonsten Fehler!).
	 *
	 * @param persistenceDir Persistenzverzeichnis
	 * @param debugLevel     Debug-Level fuer stdout
	 * @param configAuth     Konfigurationsverantwortlicher oder null (dann wird der lokale KV verwendet)
	 * @param backupModule   Sicherungsmodul ["keydox"|"filesystem"|"dvd"]
	 * @param removeOldData  gibt an, ob das uebergebene Persistenzverzeichnis geloescht werden soll
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArSWaitForParams(String persistenceDir, String debugLevel, String configAuth, String backupModule, boolean removeOldData)
			throws Exception {
		return startArSWaitForParams(persistenceDir, debugLevel, configAuth, backupModule, null, removeOldData);
	}

	/**
	 * Startet das Archivsystem. Es wird gewartet, bis die Parametrierung vollstaendig uebernommen wurde.<br> Falls im Parameter backupModule "filesystem"
	 * spezifiziert wurde, sollte ein Verzeichnis fuer die Sicherung angegeben werden.
	 *
	 * @param persistenceDir Persistenzverzeichnis
	 * @param debugLevel     Debug-Level fuer stdout
	 * @param configAuth     Konfigurationsverantwortlicher oder null (dann wird der lokale KV verwendet)
	 * @param backupModule   Sicherungsmodul ["keydox"|"filesystem"|"dvd"]
	 * @param backupDir      Sicherungsverzeichnis (kann null sein, wenn backupModule nicht "filesystem" ist)
	 * @param removeOldData  gibt an, ob das uebergebene Persistenzverzeichnis geloescht werden soll
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArSWaitForParams(
			String persistenceDir, String debugLevel, String configAuth, String backupModule, String backupDir, boolean removeOldData) throws Exception {
		ArrayList<String> paramsList = new ArrayList<String>();
		if(removeOldData) {
			if(!Util.deleteCreateNewDir(new File(persistenceDir))) return null;
		}
		// Parameter abhaengig von der Eingabe setzen
		paramsList.add("-debugLevelStdErrText=" + debugLevel);
		paramsList.add("-debugLevelFileText=" + _debugLevelFile);
		paramsList.add("-persistenzVerzeichnis=" + persistenceDir);
		paramsList.add("-sicherungsModul=" + backupModule);
		paramsList.add("-benutzer=Tester");
		paramsList.add("-authentifizierung=" + PASSWD);
		paramsList.add("-datenverteiler=" + DAV_ADDRESS);
		if(configAuth != null) paramsList.add("-konfigurationsVerantwortlicher=" + configAuth);
		if(backupDir != null) paramsList.add("-sicherungsVerzeichnis=" + backupDir);

		String[] params = new String[paramsList.size()];
		paramsList.toArray(params);

		ArchiveManager am = new ArchiveManager(params);
		EndOfParamListener eopl = am.getArchivConfig().createEndOfParamListener();
		am.run(); // Das ArS wartet von selbst auf die Archiveinstellungen
		eopl.waitForParams(); // Parametrierung durch Konfig-App abwarten
		return am;
	}

	/**
	 * Startet das Archivsystem. Es wird gewartet, bis die Parametrierung vollstaendig uebernommen wurde.<br> Falls im Parameter backupModule "filesystem"
	 * spezifiziert wurde, sollte ein Verzeichnis fuer die Sicherung angegeben werden.
	 *
	 * @param persistenceDir       Persistenzverzeichnis
	 * @param debugLevel           Debug-Level fuer stdout
	 * @param configAuth           Konfigurationsverantwortlicher oder null (dann wird der lokale KV verwendet)
	 * @param backupPropertiesFile Datei mit Backup-Einstellungen
	 * @param removeOldData        gibt an, ob das uebergebene Persistenzverzeichnis geloescht werden soll
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArSWaitForParams(
			String persistenceDir, String debugLevel, String configAuth, File backupPropertiesFile, boolean removeOldData) throws Exception {
		ArrayList<String> paramsList = new ArrayList<String>();
		if(removeOldData) {
			if(!Util.deleteCreateNewDir(new File(persistenceDir))) return null;
		}
		// Parameter abhaengig von der Eingabe setzen
		paramsList.add("-debugLevelStdErrText=" + debugLevel);
		paramsList.add("-debugLevelFileText=" + _debugLevelFile);
		paramsList.add("-persistenzVerzeichnis=" + persistenceDir);
		paramsList.add("-backup=" + backupPropertiesFile.getAbsolutePath());
		paramsList.add("-benutzer=Tester");
		paramsList.add("-authentifizierung=" + PASSWD);
		paramsList.add("-datenverteiler=" + DAV_ADDRESS);
		if(configAuth != null) paramsList.add("-konfigurationsVerantwortlicher=" + configAuth);

		String[] params = new String[paramsList.size()];
		paramsList.toArray(params);

		ArchiveManager am = new ArchiveManager(params);
		EndOfParamListener eopl = am.getArchivConfig().createEndOfParamListener();
		am.run(); // Das ArS wartet von selbst auf die Archiveinstellungen
		eopl.waitForParams(); // Parametrierung durch Konfig-App abwarten
		return am;
	}

	/**
	 * Startet das Archivsystem und wartet bis die Parametrierung bernommen wurde.
	 *
	 * @param args Kommandozeilenparameter fr das Archivsystem
	 *
	 * @return Archivmanager
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArsWaitForParamsWithArgs(String args[]) throws Exception {
		ArchiveManager am = new ArchiveManager(args);
		EndOfParamListener eopl = am.getArchivConfig().createEndOfParamListener();
		am.run(); // Das ArS wartet von selbst auf die Archiveinstellungen
		eopl.waitForParams(); // Parametrierung durch Konfig-App abwarten
		return am;
	}

	/**
	 * Loescht das uebergebene Persistenzverzeichnis und startet das Archivsystem. Es wird gewartet, bis die Parametrierung vollstaendig uebernommen wurde.
	 *
	 * @param persistenceDir Persistenzverzeichnis
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArSWaitForParamsRemoveOldData(String persistenceDir) throws Exception {
		return startArSWaitForParamsRemoveOldData(persistenceDir, "INFO");
	}

	/**
	 * Loescht das uebergebene Persistenzverzeichnis und startet das Archivsystem. Es wird gewartet, bis die Parametrierung vollstaendig uebernommen wurde.
	 *
	 * @param persistenceDir Persistenzverzeichnis
	 * @param debugLevel     Debug-Level fuer stdout
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArSWaitForParamsRemoveOldData(String persistenceDir, String debugLevel) throws Exception {
		return startArSWaitForParamsRemoveOldData(persistenceDir, debugLevel, null);
	}

	/**
	 * Loescht das uebergebene Persistenzverzeichnis und startet das Archivsystem. Es wird gewartet, bis die Parametrierung vollstaendig uebernommen wurde.
	 *
	 * @param persistenceDir Persistenzverzeichnis
	 * @param debugLevel     Debug-Level fuer stdout
	 * @param configAuth     Konfigurationsverantwortlicher oder null (dann wird der lokale KV verwendet)
	 *
	 * @return Archivsystem
	 *
	 * @throws Exception
	 */
	public static ArchiveManager startArSWaitForParamsRemoveOldData(String persistenceDir, String debugLevel, String configAuth) throws Exception {
		if(!Util.deleteCreateNewDir(new File(persistenceDir))) {
			return null;
		}
		else {
			return startArSWaitForParams(persistenceDir, debugLevel, configAuth);
		}
	}

	/** Sendet Containereinstellungen ueber den Datenverteiler. */
	public static void sendContainerSettings(String usr, String pw, String kvPid, ContainerSettings cs) throws Exception {
		connect(usr, pw);
		try {
			AttributeGroup atg = davCon.getDataModel().getAttributeGroup(ContainerSettings.ATTR_GROUP);
			DataDescription dd = new DataDescription(atg, davCon.getDataModel().getAspect(SEND_ASPECT));

			SimpleSender.sendWait(davCon, davCon.getDataModel().getObject(kvPid), dd, cs.createData(davCon), 0);
		}
		finally {
			disconnect();
		}
	}

	private static class Receiver implements ClientReceiverInterface {

		Object mutex = new Object();

		int expMsgType, expMsgIndex;

		String errMsg;

		boolean gotData = false;

		public Receiver(int expMsg, int expIdx) {
			expMsgType = expMsg;
			expMsgIndex = expIdx;
		}

		public void update(ResultData[] results) {
			for(ResultData result : results) {
				if(result.hasData()) {
					if((result.getData().getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).intValue() != expMsgType) || (
							result.getData().getUnscaledValue(PidQuery.ATT_QUERY_IDX_NAME).intValue() != expMsgIndex)) {
						return;
					}
					try {
						Deserializer ds = SerializingFactory.createDeserializer(new ByteArrayInputStream(result.getData().getUnscaledArray(PidQuery.ATT_DATA_NAME).getByteArray()));
						errMsg = (ds.readByte() == 1) ? "" : ds.readString();
						synchronized(mutex) {
							gotData = true;
							mutex.notify();
						}
					}
					catch(Exception e) {
						e.printStackTrace();
					}
				}
			}
		}
	}

	/** Sendet einen Nachforderungsauftrag, wartet und liefert den Antwortstring. */
	public static String sendRequestRequestWait(String usr, String pw, String kvPid, int requestIndex, RequestDataSpec[] rds) throws Exception {
		return sendRequestRequestWait(usr, pw, kvPid, requestIndex, rds, null);
	}

	public static String sendHeaderRestoreRequest(String usr, String pw, String kvPid, int requestIndex, int mediumID) throws Exception {
		connect(usr, pw);
		try {
			AttributeGroup atg = davCon.getDataModel().getAttributeGroup(PidQuery.ATG_QUERY_PID);
			Aspect asp = davCon.getDataModel().getAspect(PidQuery.ASP_QUERY_PID);
			Data data = davCon.createData(atg);
			data.getReferenceValue(PidQuery.ATT_SENDER_NAME).setSystemObject(davCon.getLocalApplicationObject());
			data.getUnscaledValue(PidQuery.ATT_QUERY_IDX_NAME).set(requestIndex);
			data.getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).set(MessageType.HEADER_RESTORE);

			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			Serializer ser = SerializingFactory.createSerializer(bos);
			ser.writeInt(ser.getVersion());
			ser.writeInt(mediumID);
			SerializerUtil.insertArray(data.getArray(PidQuery.ATT_DATA_NAME), bos.toByteArray(), bos.toByteArray().length);

			Receiver rec = new Receiver(MessageType.HEADER_RESTORE_RESULT, requestIndex);
			davCon.subscribeReceiver(
					rec,
					davCon.getLocalApplicationObject(),
					new DataDescription(atg, davCon.getDataModel().getAspect(PidQuery.ASP_RESPONSE_PID)),
					ReceiveOptions.normal(),
					ReceiverRole.drain()
			);

			try {
				SimpleSender.sendWait(davCon, davCon.getDataModel().getObject(kvPid), new DataDescription(atg, asp), data, 0);

				synchronized(rec.mutex) {
					while(!rec.gotData) rec.mutex.wait();
				}
			}
			finally {
				davCon.unsubscribeReceiver(rec, davCon.getLocalApplicationObject(), new DataDescription(atg, asp));
			}

			return rec.errMsg;
		}
		finally {
			disconnect();
		}
	}

	/**
	 * Sendet einen Nachforderungsauftrag, wartet und liefert den Antwortstring. Bei dieser Methode koennen die zu verwendenden Archivsysteme angegeben werden.
	 * Diese werden anstelle der parametrierten Nachforderungs-Archivsysteme verwendet.
	 */
	public static String sendRequestRequestWait(String usr, String pw, String kvPid, int requestIndex, RequestDataSpec[] rds, SystemObject[] archives)
			throws Exception {
		connect(usr, pw);
		try {
			AttributeGroup atg = davCon.getDataModel().getAttributeGroup(PidQuery.ATG_QUERY_PID);
			Aspect asp = davCon.getDataModel().getAspect(PidQuery.ASP_QUERY_PID);
			Data data = davCon.createData(atg);
			data.getReferenceValue(PidQuery.ATT_SENDER_NAME).setSystemObject(davCon.getLocalApplicationObject());
			data.getUnscaledValue(PidQuery.ATT_QUERY_IDX_NAME).set(requestIndex);
			data.getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).set(MessageType.REQUEST_DID);
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			Serializer ser = SerializingFactory.createSerializer(bos);
			ser.writeInt(ser.getVersion());

			if(rds == null || rds.length == 0) {
				ser.writeInt(0); // keine Datenspezifikation
			}
			else {
				ser.writeInt(rds.length); // Lnge der Liste der Bereiche 1
				for(RequestDataSpec r : rds) {
					ser.writeByte(r.getTT());
					ser.writeByte(r.relative ? 1 : 0);
					ser.writeLong(r.start); // start
					ser.writeLong(r.end); // ende
					ser.writeByte(r.adkc[0] ? 1 : 0); // OA
					ser.writeByte(r.adkc[1] ? 1 : 0); // ON
					ser.writeByte(r.adkc[2] ? 1 : 0); // NA
					ser.writeByte(r.adkc[3] ? 1 : 0); // NN
					ser.writeInt(r.order); // ArchivOrder
					ser.writeInt(r.requestOption); // ArchiveRequestOption
					ser.writeObjectReference(davCon.getDataModel().getObject(r.atg));
					ser.writeObjectReference(davCon.getDataModel().getObject(r.asp)); // dd
					ser.writeShort(r.sv); // SV
					ser.writeObjectReference(davCon.getDataModel().getObject(r.obj)); // Systemobject
				}
			}
			int anzArS = archives != null ? archives.length : 0;
			ser.writeInt(anzArS); // keine Archive
			for(int i = 0; i < anzArS; i++) {
				ser.writeObjectReference(archives[i]);
			}

			SerializerUtil.insertArray(data.getArray(PidQuery.ATT_DATA_NAME), bos.toByteArray(), bos.toByteArray().length);

			Receiver rec = new Receiver(MessageType.REQUEST_DID_RESULT, requestIndex);
			davCon.subscribeReceiver(
					rec,
					davCon.getLocalApplicationObject(),
					new DataDescription(atg, davCon.getDataModel().getAspect(PidQuery.ASP_RESPONSE_PID)),
					ReceiveOptions.normal(),
					ReceiverRole.drain()
			);

			try {
				SimpleSender.sendWait(davCon, davCon.getDataModel().getObject(kvPid), new DataDescription(atg, asp), data, 0);

				synchronized(rec.mutex) {
					while(!rec.gotData) rec.mutex.wait();
				}
			}
			finally {
				davCon.unsubscribeReceiver(rec, davCon.getLocalApplicationObject(), new DataDescription(atg, asp));
			}

			return rec.errMsg;
		}
		finally {
			disconnect();
		}
	}

	/**
	 * Sendet eine Archivinformationsanfrage an das Archivsystem. Stellt dazu eine eigene Verbindung zum Datenverteiler her.
	 *
	 * @param usr   Dav-Benutzer, z.B. "Tester".
	 * @param pw    Dav-Passwort, z.B. "geheim".
	 * @param kvPid Pid des Konfigurationsverantwortlichen, z.B. "kv.testKonfiguration".
	 * @param ads   Parameter der Anfrage.
	 *
	 * @return Ergebnis der Anfrage
	 *
	 * @throws Exception Fehler ist aufgetreten.
	 */
	public static ArchiveInfoQueryResult sendArchiveInfoRequestWait(String usr, String pw, String kvPid, ArchiveDataSpecification[] ads) throws Exception {
		connect(usr, pw);
		try {
			List<ArchiveDataSpecification> adsList = new ArrayList<ArchiveDataSpecification>();
			for(int i = 0; i < ads.length; i++) {
				adsList.add(ads[i]);
			}
			ArchiveInfoQueryResult result = davCon.getArchive().requestInfo(adsList);
			// Vor dem Beenden der Verbindung!!!! Benoetigt den DAV !!!
			result.isRequestSuccessful();
			return result;
		}
		finally {
			disconnect();
		}
	}

	/**
	 * Sendet Containereinstellungen ueber den Datenverteiler und wartet bis das uebergeben Archivsystem darauf reagiert hat.
	 *
	 * @throws InterruptedException
	 */
	public static void sendContainerSettingsWait(String usr, String pw, String kvPid, ContainerSettings cs, ArchiveManager aMgr) throws Exception {
		ArSSettingListener eosl = aMgr.getInQueuesMgr().getArchiveSettingsTask().createSettingLstnr(ArchiveSettingsTask.ARCHIVCONTAINER);
		sendContainerSettings(usr, pw, kvPid, cs);
		eosl.waitForSettings();
	}

	public static void sendQuickContainerSettingsWait(String usr, String pw, String kvPid, int maxDS, int maxSize, long maxTime, ArchiveManager aMgr)
			throws Exception {
		ContainerSettings cs = new ContainerSettings();
		cs.stdCloseConditions.maxContAnzDS = maxDS;
		cs.stdCloseConditions.maxContSize = maxSize;
		cs.stdCloseConditions.maxContTime = maxTime;
		sendContainerSettingsWait(usr, pw, kvPid, cs, aMgr);
	}

	/**
	 * Setzt atg.archivContainer
	 *
	 * @param usr     Benutzer
	 * @param pw      Passwort
	 * @param kvPid   Konfigurationsverantwortlicher
	 * @param maxDS   max. Anzahl Datensaetze pro Container
	 * @param maxSize max. Groesse eines Containers in Bytes
	 * @param maxTime max. Zeitspanne eines Containers in Sekunden
	 *
	 * @throws Exception
	 */
	public static void sendQuickContainerSettings(String usr, String pw, String kvPid, int maxDS, int maxSize, long maxTime) throws Exception {
		ContainerSettings cs = new ContainerSettings();
		cs.stdCloseConditions.maxContAnzDS = maxDS;
		cs.stdCloseConditions.maxContSize = maxSize;
		cs.stdCloseConditions.maxContTime = maxTime;
		sendContainerSettings(usr, pw, kvPid, cs);
	}

	/** Sendet Archiveinstellungen ueber den Datenverteiler. */
	public static void sendArSSettings(String usr, String pw, String kvPid, ArchiveSettings einst) throws Exception {
		connect(usr, pw);
		try {
			DataDescription dd = new DataDescription(
					davCon.getDataModel().getAttributeGroup(ArchiveSettings.ATTR_GROUP), davCon.getDataModel().getAspect(SEND_ASPECT)
			);

			SimpleSender.sendWait(davCon, davCon.getDataModel().getObject(kvPid), dd, einst.createData(davCon), 0);
		}
		finally {
			disconnect();
		}
	}

	/** Sendet Archiveinstellungen ueber den Datenverteiler und wartet bis das uebergeben Archivsystem darauf reagiert hat. */
	public static void sendArSSettingsWait(String usr, String pw, String kvPid, ArchiveSettings einst, ArchiveManager aMgr) throws Exception {
		ArSSettingListener eosl = aMgr.getInQueuesMgr().getArchiveSettingsTask().createSettingLstnr(ArchiveSettingsTask.ARCHIVEINST);
		sendArSSettings(usr, pw, kvPid, einst);
		eosl.waitForSettings();
	}

	/**
	 * Parametriert das Archivsystem unter dem angegebenen Konfigurationsverantwortlichen anhand der uebergebenen Parameter.
	 *
	 * @param kvPid Konfigurationsverantwortlicher
	 * @param ps    Parametersatz
	 *
	 * @throws ConfigurationException
	 * @throws MissingParameterException
	 * @throws CommunicationError
	 * @throws ConnectionException
	 * @throws InconsistentLoginException
	 * @throws SenderException
	 */
	public static void sendArsParams(String usr, String pw, String kvPid, ParamSatz[] ps) throws Exception {
		connect(usr, pw);
		try {
			AttributeGroup atg = davCon.getDataModel().getAttributeGroup("atg.archiv");
			Data data = davCon.createData(atg);

			data.getArray("ParameterSatz").setLength(ps.length);
			for(int p = 0; p < ps.length; p++) {
				data.getArray("ParameterSatz").getItem(p).getArray("Bereich").setLength(ps[p].ber.length);
				for(int i = 0; i < ps[p].ber.length; i++) {
					data.getArray("ParameterSatz").getItem(p).getArray("Bereich").getItem(i).asTextValue().setText(ps[p].ber[i]);
				}

				// Datenspezifikationen
				for(int d = 0; d < ps[p].ds.length; d++) {
					data.getArray("ParameterSatz").getItem(p).getArray("DatenSpezifikation").setLength(ps[p].ds.length);
					data.getArray("ParameterSatz").getItem(p).getArray("DatenSpezifikation").getItem(d).getArray("Objekt").setLength(ps[p].ds[d].objs.length);
					for(int i = 0; i < ps[p].ds[d].objs.length; i++) {
						data.getArray("ParameterSatz").getItem(p).getArray("DatenSpezifikation").getItem(d).getArray("Objekt").getItem(i).asTextValue().setText(
								ps[p].ds[d].objs[i]
						);
					}
					data.getArray("ParameterSatz").getItem(p).getArray("DatenSpezifikation").getItem(d).getArray("AttributGruppe").setLength(ps[p].ds[d].atgs.length);
					for(int i = 0; i < ps[p].ds[d].atgs.length; i++) {
						data.getArray("ParameterSatz").getItem(p).getArray("DatenSpezifikation").getItem(d).getArray("AttributGruppe").getItem(i).asTextValue().setText(
								ps[p].ds[d].atgs[i]
						);
					}
					data.getArray("ParameterSatz").getItem(p).getArray("DatenSpezifikation").getItem(d).getArray("Aspekt").setLength(ps[p].ds[d].asps.length);
					for(int i = 0; i < ps[p].ds[d].asps.length; i++) {
						data.getArray("ParameterSatz").getItem(p).getArray("DatenSpezifikation").getItem(d).getArray("Aspekt").getItem(i).asTextValue().setText(
								ps[p].ds[d].asps[i]
						);
					}
					data.getArray("ParameterSatz").getItem(p).getArray("DatenSpezifikation").getItem(d).getItem("SimulationsVariante").asUnscaledValue().set(ps[p].ds[d].sv);
				}

				// Einstellungen
				data.getArray("ParameterSatz").getItem(p).getItem("Einstellungen").getItem("Archivieren").asUnscaledValue().set(
						ps[p].es.arch ? 1 : 0
				);
				int nfLen = ps[p].es.nachf != null ? ps[p].es.nachf.length : 0;
				data.getArray("ParameterSatz").getItem(p).getItem("Einstellungen").getArray("Nachfordern").setLength(nfLen);
				for(int i = 0; i < nfLen; i++) {
					data.getArray("ParameterSatz").getItem(p).getItem("Einstellungen").getArray("Nachfordern").getItem(i).asTextValue().setText(ps[p].es.nachf[i]);
				}
				data.getArray("ParameterSatz").getItem(p).getItem("Einstellungen").getItem("Sichern").asUnscaledValue().set(
						ps[p].es.sich ? 1 : 0
				);
				data.getArray("ParameterSatz").getItem(p).getItem("Einstellungen").getArray("Quittieren").setLength(
						ps[p].es.quitt != null ? 1 : 0
				);
				if(ps[p].es.quitt != null) {
					data.getArray("ParameterSatz").getItem(p).getItem("Einstellungen").getArray("Quittieren").getItem(0).asTextValue().setText(ps[p].es.quitt);
				}
				data.getArray("ParameterSatz").getItem(p).getItem("Einstellungen").getItem("Vorhalten").asTimeValue().setSeconds(ps[p].es.vorh);
			}

			SimpleSender.sendWait(
					davCon, davCon.getDataModel().getObject(kvPid), new DataDescription(atg, davCon.getDataModel().getAspect(SEND_ASPECT)), data, 0
			);
		}
		finally {
			disconnect();
		}
	}

	/** Sendet einen Parametersatz zum DAV und wartet bis das uebergebene Archivsystem darauf reagiert hat. */
	public static void sendArsParamsWait(String usr, String pw, ParamSatz[] ps, ArchiveManager aMgr) throws Exception {
		EndOfParamListener eopl = aMgr.getArchivConfig().createEndOfParamListener();
		sendArsParams(usr, pw, aMgr.getConfigAuth().getPid(), ps);
		eopl.waitForParams(); // Parametrierung abwarten
	}

	private static void connect(String usr, String pw) throws Exception {
		suspendDebug();
		davParams = new ClientDavParameters(
				new ArgumentList(
						new String[]{
								"-debugLevelStdErrText=INFO", "-datenverteiler=" + DAV_ADDRESS
						}
				)
		);
		davParams.setApplicationName("Archiv-Parametrierer");
		davCon = new ClientDavConnection(davParams);
		davCon.connect();
		davCon.login(usr, pw);
		resumeDebug();
	}

	private static void disconnect() {
		davCon.disconnect(false, "");
		davCon = null;
	}

	// *****************************************************************************************************************************
	// Hilfsklassen:

	public static class Datenspez {

		String[] objs, atgs, asps;

		int sv;

		public Datenspez(String[] objs, String[] atgs, String[] asps, int sv) {
			this.objs = objs;
			this.atgs = atgs;
			this.asps = asps;
			this.sv = sv;
		}

		public Datenspez(List<String> objs, String atg, String asp) {
			this.objs = new String[objs.size()];
			for(int i = 0; i < objs.size(); i++) {
				this.objs[i] = objs.get(i);
			}
			this.atgs = new String[]{atg};
			this.asps = new String[]{asp};
			this.sv = 0;
		}

		public Datenspez(String[] objs, String atg, String asp) {
			this.objs = objs;
			this.atgs = new String[]{atg};
			this.asps = new String[]{asp};
			this.sv = 0;
		}

		public Datenspez(String obj, String atg, String asp, int sv) {
			this.objs = new String[]{obj};
			this.atgs = new String[]{atg};
			this.asps = new String[]{asp};
			this.sv = sv;
		}

		public Datenspez(String obj, String atg, String asp) {
			this.objs = new String[]{obj};
			this.atgs = new String[]{atg};
			this.asps = new String[]{asp};
			this.sv = 0;
		}

		public Datenspez(String atg, String asp) {
			this.objs = new String[0];
			this.atgs = new String[]{atg};
			this.asps = new String[]{asp};
			this.sv = 0;
		}
	}

	public static class Einst {

		public boolean arch, sich;

		public String[] nachf;

		public String quitt;

		public int vorh;

		/** archivieren, nicht nachfordern, nicht sichern, nicht quittieren, unendlich lang vorhalten */
		public static Einst getStdArchEinst() {
			return new Einst(true, null, false, null, Integer.MAX_VALUE);
		}

		/** archivieren, nicht nachfordern, sichern, nicht quittieren, unendlich lang vorhalten */
		public static Einst getArchSichEinst() {
			return new Einst(true, null, true, null, Integer.MAX_VALUE);
		}

		public Einst(boolean archivieren, String[] nachfordern, boolean sichern, String quittieren, int vorhalten) {
			arch = archivieren;
			nachf = (nachfordern != null) ? nachfordern : new String[0];
			quitt = quittieren;
			sich = sichern;
			vorh = vorhalten;
		}
	}

	public static class ParamSatz {

		public static ParamSatz[] EMPTY_PARAMS = new ParamSatz[]{};

		public String[] ber;

		public Datenspez[] ds;

		public Einst es;

		public ParamSatz(String[] bereich, Datenspez[] datenspezifikationen, Einst einstellungen) {
			ber = bereich;
			ds = datenspezifikationen;
			es = einstellungen;
		}

		public ParamSatz(String[] bereich, Datenspez datenspezifikation, Einst einstellungen) {
			ber = bereich;
			ds = new Datenspez[]{datenspezifikation};
			es = einstellungen;
		}
	}

	public static class RequestDataSpec {

		TimingType tt;

		boolean relative;

		long start, end;

		boolean[] adkc;

		int order, requestOption;

		long obj, atg, asp;

		short sv;

		public RequestDataSpec(TimingType tt, long start, long end, long obj, long atg, long asp, int sv) {
			this.tt = tt;
			this.relative = false;
			this.start = start;
			this.end = end;
			adkc = new boolean[]{true, true, true, true};
			this.order = ArchiveOrder.BY_INDEX.getCode();
			this.requestOption = ArchiveRequestOption.NORMAL.getCode();
			this.obj = obj;
			this.atg = atg;
			this.asp = asp;
			this.sv = (short)sv;
		}

		public RequestDataSpec(
				TimingType tt,
				boolean relative,
				long start,
				long end,
				boolean oa,
				boolean on,
				boolean na,
				boolean nn,
				int order,
				int requestOption,
				long obj,
				long atg,
				long asp,
				short sv) {
			this.tt = tt;
			this.relative = relative;
			this.start = start;
			this.end = end;
			adkc = new boolean[]{oa, on, na, nn};
			this.order = order;
			this.requestOption = requestOption;
			this.obj = obj;
			this.atg = atg;
			this.asp = asp;
			this.sv = (short)sv;
		}

		public int getTT() {
			if(tt.equals(TimingType.DATA_INDEX)) {
				return 3;
			}
			else if(tt.equals(TimingType.ARCHIVE_TIME)) {
				return 2;
			}
			else if(tt.equals(TimingType.DATA_TIME)) {
				return 1;
			}
			else {
				return 0;
			}
		}
	}
}
